@@ -3,6 +3,8 @@ Contains the Unfold class, to extract datapackage files.
 
 """
 import copy
+import os
+import pickle
 import uuid
 from ast import literal_eval
 from collections import defaultdict
@@ -33,12 +35,23 @@ from .data_cleaning import (
     correct_fields_format,
     get_list_of_unique_datasets,
     get_outdated_flows,
+    get_outdated_units,
     remove_categories_for_technosphere_flows,
     remove_missing_fields,
 )
-from .export import UnfoldExporter
+
+try:
+    import bw_processing
+
+    from .brightway25_export import write_brightway_database
+
+except ImportError:
+    from .brightway2_export import write_brightway_database
+
 from .utils import HiddenPrints
 
+DIR_CACHED_DB = Path(__file__).parent / "cached_databases"
+
 
 def _c(value):
     """Converts zero to one."""
@@ -88,6 +101,42 @@ def get_list_unique_exchanges(databases: list) -> list:
     )
 
 
+def check_cached_database(name) -> list:
+    """
+    Checks if a cached database exists, and if not, extracts it.
+    :param name: name of the database
+    :return: the database
+    """
+
+    # check that directory exists, otherwise create it
+    Path(DIR_CACHED_DB).mkdir(parents=True, exist_ok=True)
+    bw2io_version = "".join(tuple(map(str, bw2io.__version__)))
+    file_name = Path(DIR_CACHED_DB / f"cached_{name}_{bw2io_version}.pickle")
+
+    # check that file path leads to an existing file
+    if file_name.exists() and len(bw2data.Database(name)) > 0:
+        # check that the database has same length as the cached one
+        database = pickle.load(open(file_name, "rb"))
+        if len(database) == len(bw2data.Database(name)):
+            # return the cached database
+            return pickle.load(open(file_name, "rb"))
+
+    # extract the database, pickle it for next time and return it
+    print("Cannot find cached database. Will create one now for next time...")
+    database = extract_brightway2_databases(name)
+    pickle.dump(database, open(file_name, "wb"))
+    return database
+
+
+def clear_cache():
+    """Clears the cache folder, except for files which contain __version__ in name.
+    Useful when updating `premise`
+    or encountering issues with
+    inventories.
+    """
+    [f.unlink() for f in Path(DIR_CACHED_DB).glob("*")]
+
+
 class Unfold:
     """Extracts datapackage files."""
 
@@ -96,6 +145,7 @@ class Unfold:
         self.acts_indices = None
         self.reversed_acts_indices = None
         self.outdated_flows = get_outdated_flows()
+        self.outdated_units = get_outdated_units()
         self.path = path
         self.package = Package(self.path)
         self.dependencies = self.package.descriptor["dependencies"]
@@ -254,9 +304,10 @@ class Unfold:
         """
 
         for dependency in self.dependencies:
-            database = extract_brightway2_databases(dependency["source"])
+            database = check_cached_database(dependency["source"])
 
             self.build_mapping_for_dependencies(database)
+
             if dependency.get("type") == "source":
                 self.database.extend(database)
 
@@ -433,13 +484,42 @@ class Unfold:
                 return key[0] + ", in ground", key[1], key[2], key[3]
 
             else:
-                print("Could not find key", key)
-                return key
+                print(f"Could not find key: {key}")
+                # return key
 
     def fix_key(self, key: tuple) -> tuple:
         if key in self.dependency_mapping:
             return self.dependency_mapping[key]
         else:
+            # try first by searching for alternative units
+            if key[4] in self.outdated_units:
+                print(
+                    f"found alternative unit: {key[4]}. Changing to {self.outdated_units[key[4]]}"
+                )
+                if (
+                    key[0],
+                    key[1],
+                    key[2],
+                    key[3],
+                    self.outdated_units[key[4]],
+                    key[5],
+                ) in self.dependency_mapping:
+                    return self.dependency_mapping[
+                        (
+                            key[0],
+                            key[1],
+                            key[2],
+                            key[3],
+                            self.outdated_units[key[4]],
+                            key[5],
+                        )
+                    ]
+
+            correct_id = self.find_correct_id(key)
+            if correct_id is None:
+                print(f"Could not find key: {key}")
+                return key
+
             return self.dependency_mapping[self.find_correct_id(key)]
 
     def get_exchange(
@@ -449,7 +529,7 @@ class Unfold:
         scenario_name: str = None,
     ):
         """
-        Return an exchange in teh form of a dictionary.
+        Return an exchange in the form of a dictionary.
         If it has a value for "categories", we inder it is a biosphere flow.
         If not, it is either a technosphere or production flow.
 
@@ -569,6 +649,19 @@ class Unfold:
                 None,
             )
 
+            if consumer_idx is None:
+                consumer_idx = self.reversed_acts_indices.get(
+                    (
+                        c_name,
+                        c_prod,
+                        None,
+                        c_loc,
+                        self.outdated_units.get(c_unit, c_unit),
+                        "production",
+                    ),
+                    None,
+                )
+
             # Look up the index of the supplier activity in the reversed activities index.
             supplier_id = (
                 s_name,
@@ -583,6 +676,19 @@ class Unfold:
                 None,
             )
 
+            if supplier_idx is None:
+                supplier_idx = self.reversed_acts_indices.get(
+                    (
+                        s_name,
+                        s_prod,
+                        s_cat,
+                        s_loc,
+                        self.outdated_units.get(s_unit, s_unit),
+                        s_type,
+                    ),
+                    None,
+                )
+
             # Multiply the appropriate element of the matrix by the scaling factor for the given scenario.
             # Use the lambda function defined above to avoid multiplying by zero.
             if supplier_idx is not None and consumer_idx is not None:
@@ -761,19 +867,40 @@ class Unfold:
         """
 
         # Get the list of datasets not used in the current scenario.
-        df_gr = self.scenario_df.groupby("to activity name").sum(numeric_only=True)
+        df_gr = self.scenario_df.groupby(
+            ["to activity name", "to reference product", "to location"]
+        ).sum(numeric_only=True)
 
         datasets_not_in_scenario = df_gr.loc[
             (df_gr[scenario_name] == 0) & (df_gr.sum(1) != 0), :
         ].index.tolist()
 
+        # remove items from datasets_not_in_scenario that are
+        # used in exchanges of database
+        for act in database:
+            for exc in act["exchanges"]:
+                if exc["type"] == "technosphere":
+                    key = (exc["name"], exc["product"], exc["location"])
+                    if key in datasets_not_in_scenario:
+                        datasets_not_in_scenario.remove(key)
+
         # Remove datasets that are not in the current scenario.
         database = [
-            act for act in database if act["name"] not in datasets_not_in_scenario
+            act
+            for act in database
+            if (act["name"], act["reference product"], act["location"])
+            not in datasets_not_in_scenario
         ]
 
         return database
 
+    def check_usage(
+        self, name: str, product: str, location: str, database: list
+    ) -> bool:
+        # check that name, product, location are not used in any exchange of database
+
+        pass
+
     def generate_superstructure_database(self) -> List[dict]:
         """
         Generates the superstructure database.
@@ -862,12 +989,9 @@ class Unfold:
         filepath = self.package.get_resource("scenario_data").source
         # ensure that all slashes in filepath are forward slashes
         filepath = filepath.replace("\\", "/")
-        self.scenario_df = pd.read_csv(filepath, keep_default_na=False, na_values="")
-
-        # Remove rows corresponding to production flows.
-        self.scenario_df = self.scenario_df.loc[
-            (self.scenario_df["flow type"] != "production")
-        ]
+        self.scenario_df = pd.read_csv(
+            filepath, keep_default_na=False, na_values="", encoding="utf-8-sig"
+        )
 
         # Drop columns corresponding to scenarios that were removed.
         self.scenario_df = self.scenario_df.drop(scenarios_to_leave_out, axis=1)
@@ -926,26 +1050,42 @@ class Unfold:
         matrix = self.populate_sparse_matrix()
 
         # Define a lambda function to replace zero values with 1.0
-        _ = lambda x: x if x != 0 else 1.0
+        _ = lambda x: 1.0 if x == 0.0 else x
 
         # Loop through each flow in self.factors and update the corresponding factor values
         for flow_id, factor in self.factors.items():
             c_name, c_prod, c_loc, c_unit = list(flow_id)[:4]
             s_name, s_prod, s_loc, s_cat, s_unit, s_type = list(flow_id)[4:]
 
-            # Get the indices of the consumer and supplier activities in the reversed_act_indices dictionary
-            consumer_idx = self.reversed_acts_indices[
-                (
+            # Get the indices of the consumer and supplier activities
+            # in the reversed_act_indices dictionary
+            consumer_key = (
+                c_name,
+                c_prod,
+                None,
+                c_loc,
+                c_unit,
+                "production",
+            )
+            if consumer_key in self.reversed_acts_indices:
+                consumer_idx = self.reversed_acts_indices[consumer_key]
+            else:
+                consumer_key = (
                     c_name,
                     c_prod,
                     None,
                     c_loc,
-                    c_unit,
+                    self.outdated_units.get(c_unit, c_unit),
                     "production",
                 )
-            ]
 
-            supplier_id = (
+                if consumer_key in self.reversed_acts_indices:
+                    consumer_idx = self.reversed_acts_indices[consumer_key]
+                else:
+                    print("not found key for consumer ", consumer_key)
+                    consumer_idx = 0
+
+            supplier_key = (
                 s_name,
                 s_prod,
                 s_cat,
@@ -954,7 +1094,7 @@ class Unfold:
                 s_type,
             )
 
-            if supplier_id not in self.reversed_acts_indices:
+            if supplier_key not in self.reversed_acts_indices:
                 try:
                     (
                         s_name,
@@ -963,7 +1103,7 @@ class Unfold:
                         s_cat,
                     ) = self.find_correct_id((s_name, s_prod, s_loc, s_cat))
 
-                    supplier_id = (
+                    supplier_key = (
                         s_name,
                         s_prod,
                         s_cat,
@@ -972,13 +1112,25 @@ class Unfold:
                         s_type,
                     )
                 except TypeError:
-                    print("not found key for", supplier_id)
+                    supplier_key = (
+                        s_name,
+                        s_prod,
+                        s_cat,
+                        s_loc,
+                        self.outdated_units.get(s_unit, s_unit),
+                        s_type,
+                    )
 
-            supplier_idx = self.reversed_acts_indices[supplier_id]
+                    if supplier_key not in self.reversed_acts_indices:
+                        print("not found key for supplier", supplier_key)
+
+            supplier_idx = self.reversed_acts_indices[supplier_key]
+
+            # Update the factor values for each scenario
+            # by multiplying with the corresponding matrix value
 
-            # Update the factor values for each scenario by multiplying with the corresponding matrix value
             for scenario, val in factor.items():
-                factor[scenario] = val * _(matrix[consumer_idx, supplier_idx])
+                factor[scenario] = val * _(matrix[supplier_idx, consumer_idx])
 
         # Create a new scenario dataframe from the updated factors dictionary
         self.scenario_df = pd.DataFrame.from_dict(self.factors).T.reset_index()
@@ -1039,9 +1191,24 @@ class Unfold:
         )
 
         # Add "from database" column based on flow type
+        # set dtype of "from database" and "to database" to string
+
+        # create "from database" column
+        self.scenario_df.loc[:, "from database"] = None
+
+        self.scenario_df = self.scenario_df.astype(
+            {"from database": "string"},
+            {"to database": "string"},
+        )
+
         self.scenario_df.loc[
-            (self.scenario_df["flow type"] == "technosphere"), "from database"
-        ] = (self.name or self.package.descriptor["name"])
+            (self.scenario_df["flow type"] == "technosphere")
+            | (self.scenario_df["flow type"] == "production"),
+            "from database",
+        ] = (
+            self.name or self.package.descriptor["name"]
+        )
+
         self.scenario_df.loc[
             (self.scenario_df["flow type"] == "biosphere"), "from database"
         ] = "biosphere3"
@@ -1069,6 +1236,7 @@ class Unfold:
         scenarios: List[int] = None,
         dependencies: dict = None,
         superstructure: bool = False,
+        export_dir: str = None,
         name: str = None,
     ):
         """
@@ -1078,6 +1246,7 @@ class Unfold:
         :param dependencies: A dictionary containing additional inventory databases that may be needed for extraction. Default is None.
         :param superstructure: A boolean indicating whether to generate a scenario difference file and a superstructure database. Default is False.
         :param name: A string indicating the name of the superstructure database. Default is None.
+        :param export_dir: A string indicating the directory to which the scenario difference file should be exported. Default is current working directory.
         :return: None
 
         Behavior:
@@ -1097,7 +1266,6 @@ class Unfold:
         self.check_dependencies(dependencies)
         self.extract_source_database()
         self.extract_additional_inventories()
-
         self.format_dataframe(scenarios=scenarios, superstructure=superstructure)
         self.generate_factors()
 
@@ -1114,48 +1282,54 @@ class Unfold:
             self.format_superstructure_dataframe()
             self.database = self.generate_superstructure_database()
 
-        self.write(superstructure=superstructure)
+        self.write(superstructure=superstructure, export_dir=export_dir)
 
-    def write(self, superstructure: bool = False):
+    def write(self, superstructure: bool = False, export_dir: str = None):
         """
         Write the databases.
         If superstructure is True, write the scenario difference file,
         along with a database.
 
         :param superstructure: bool, default False
+        :param export_dir: str, default current working directory
 
         """
 
         if not superstructure:
             for scenario, database in self.databases_to_export.items():
                 change_db_name(data=database, name=scenario)
-                # check_exchanges_input(database, self.dependency_mapping)
                 link_internal(database)
                 check_internal_linking(database)
                 check_duplicate_codes(database)
                 correct_fields_format(database, scenario)
                 print(f"Writing database for scenario {scenario}...")
-                UnfoldExporter(scenario, database).write_database()
+                write_brightway_database(data=database, name=scenario)
 
         else:
-            try:
-                self.scenario_df.to_excel(
-                    f"{self.name or self.package.descriptor['name']}.xlsx", index=False
-                )
-            except ValueError:
-                # from https://stackoverflow.com/questions/66356152/splitting-a-dataframe-into-multiple-sheets
-                GROUP_LENGTH = 1000000  # set nr of rows to slice df
-                with pd.ExcelWriter(
-                    f"{self.package.descriptor['name']}.xlsx"
-                ) as writer:
-                    for i in range(0, len(self.scenario_df), GROUP_LENGTH):
-                        self.scenario_df[i : i + GROUP_LENGTH].to_excel(
-                            writer, sheet_name=f"Row {i}", index=False, header=True
-                        )
+            source_db = [db for db in self.dependencies if db.get("type") == "source"]
 
-            print(
-                f"Scenario difference file exported to {self.package.descriptor['name']}.xlsx!"
+            if len(source_db) == 1:
+                source_db = source_db[0]
+            else:
+                source_db = {"name": "unknown", "version": "unknown"}
+
+            if self.name:
+                filename = f"SDF {source_db['name']} {self.name}.csv"
+            else:
+                filename = f"SDF {source_db['name']} {source_db['version']} {self.package.descriptor['name']}.csv"
+
+            if export_dir is None:
+                export_dir = os.getcwd()
+
+            filename = os.path.join(export_dir, filename)
+
+            self.scenario_df.to_csv(
+                filename,
+                index=False,
+                encoding="utf-8-sig",
             )
+
+            print(f"Scenario difference file exported to {filename}")
             print("")
             print("Writing superstructure database...")
             change_db_name(self.database, self.name or self.package.descriptor["name"])
@@ -1168,6 +1342,6 @@ class Unfold:
             correct_fields_format(
                 self.database, self.name or self.package.descriptor["name"]
             )
-            UnfoldExporter(
-                self.name or self.package.descriptor["name"], self.database
-            ).write_database()
+            write_brightway_database(
+                data=self.database, name=self.name or self.package.descriptor["name"]
+            )
